// J-INTERCAL 0.12 Copyright (c) 2000-2001 Karsten Sperling, Karlsruhe, Germany.
// This software is released under the GNU LGPL, see LICENSE for details.

package intercal;

import java.io.*;
import java.util.Vector;
import java.util.Hashtable;

public class Compiler {

  protected final static int VERSION        = 1;
  protected final static String JICKVERSION = "0.12";

  protected final static int INFO_ABSTAIN   = 0x8000;
  protected final static int INFO_STASH     = 0x8000;
  protected final static int INFO_CMDSIZE   = 3;
  protected final static int INFO_DELTANEXT = 4;

  protected final static int SIZE_CMDHEADER = 5;
  protected final static int SIZE_COMEFROM  = 3;
  protected final static int LINK_NEXT      = SIZE_CMDHEADER + 1;
  protected final static int LINK_ABSTAIN   = SIZE_CMDHEADER + 2;

  protected final static int VAR_SCALAR16   = 0;
  protected final static int VAR_SCALAR32   = 1;
  protected final static int VAR_ARRAY16    = 2;
  protected final static int VAR_ARRAY32    = 3;
  
  protected final static int UNOP_AND       = 1;
  protected final static int UNOP_OR        = 2;
  protected final static int UNOP_XOR       = 3;
  
  protected final static int CMD_CALCULATE  = 0;
  protected final static int CMD_NEXT       = 1;
  protected final static int CMD_RESUME     = 2;
  protected final static int CMD_FORGET     = 3;
  protected final static int CMD_STASH      = 4;
  protected final static int CMD_RETRIEVE   = 5;
  protected final static int CMD_IGNORE     = 6;
  protected final static int CMD_REMEMBER   = 7;
  protected final static int CMD_ABSTAIN    = 8;
  protected final static int CMD_REINSTATE  = 9;
  protected final static int CMD_READOUT    = 10;
  protected final static int CMD_WRITEIN    = 11;
  protected final static int CMD_COMEFROM   = 12;
  protected final static int CMD_SPLAT      = 13;
  protected final static int CMD_GIVEUP     = 14;

  protected final static int CMD_NUMCMDMAP  = CMD_COMEFROM + 1;
  protected final static int CMD_CANABSTAIN = CMD_SPLAT;
  
  protected final static String SYSL_FILENAME  = "syslib.i";
  protected final static int SYSL_LABELMIN  = 1000;
  protected final static int SYSL_LABELMAX  = 1999;
  protected final static int SYSL_NOINCLUDE = -1;
  protected final static int SYSL_INCLUDE   = 1;
  

  /////////////////////////////////////////////////////////////////////////////
  // static methods
  //
  public static void main(String argv[]) throws IOException {
    Compiler c = new Compiler();
    File syslib = new File(SYSL_FILENAME);
    int i, files = 0;
    boolean e774 = true;

    try {
      for (i=0; i<argv.length; i++) {
        if ((argv[i].length() == 2) && (argv[i].charAt(0) == '-')) {
          switch (argv[i].charAt(1)) {
            case '?':
              System.err.println(
                "J-INTERCAL " + JICKVERSION + " JIVM step " + VERSION +
                        ". Portable INTERCAL to Java Compiler for Java 1.1+\n" +
                "Usage: jick [options] <source> ...\n\n" +
                "  -?\t\tprint this help text. See README for additional information.\n" +
                "  -b\t\treduce the probability of E774 to zero.\n" +
                "  -l <filename>\tsyslib file for automagic inclusion .(default: ./syslib.i)\n");
              return;
            
            case 'b':
              e774 = false;
              break;
              
            case 'l':
              if (i+1 >= argv.length) fail(601);
              syslib = new File(argv[++i]);
              break;
              
            default:
              fail(601);
          }
        }
        else {
          if (e774 && (Math.random() < 0.1)) fail(774);
          c.compile(new File(argv[i]), syslib);
          files++;
        }
      }
      if (files == 0) fail(777);
    }
    catch (IntercalException e) {
      System.err.println(e);
    }
  }


  /////////////////////////////////////////////////////////////////////////////
  // instance methods and data
  //
  protected class Statement {
    public int line, label, id, offset, realoffset;
    public StringBuffer asm;

    public Statement comefrom, target;

    public Statement() {
      asm = new StringBuffer(128);
    }
  }

  protected Vector code;
  protected char[] text;
  protected int ti, line, usesyslib;
  
  protected Hashtable labels;
  protected boolean expr32;
  protected int[] maxvar;
  
  public void compile(File source, File syslib) throws IntercalException, IOException {
    String name;
    File target;

    // figure out the target file
    {
      int pos;

      if (!source.canRead()) fail(777);

      name = source.getName();
      pos = name.lastIndexOf('.');
      if ((pos < 0) || !name.substring(pos).toLowerCase().equals(".i")) fail(988);
      name = name.substring(0, pos);

      target = new File(source.getParent(), name + ".class");
      target.delete();
    }

    // set up a few globals
    code   = new Vector(512, 256);
    labels = new Hashtable(1024, 0.7f);
    maxvar = new int[4];
    maxvar[0] = maxvar[1] = maxvar[2] = maxvar[3] = 0;
    usesyslib = 0;
    
    compLoad(source);                 // parse the input file
    compParse();
    if (usesyslib == SYSL_INCLUDE) {  // automagic syslib inclusion
      if ((syslib == null) || !syslib.canRead()) fail(127);
      compLoad(syslib);
      compParse();
    }
    compResolve();                    // resolve labels into statements
    compFlatten();                    // calculate flat addresses
    compLink();                       // resolve statements into addresses
    compOutput(name, target);         // write class file
    
    code    = null;
    labels  = null;
    maxvar  = null;
  }

  protected void compLoad(File source) throws IOException {
    BufferedReader inr = new BufferedReader(new InputStreamReader(new FileInputStream(source)));
    StringBuffer buf = new StringBuffer();
    String line;
  
    while ((line = inr.readLine()) != null) {
      buf.append(line);
      buf.append('\n');
    }
  
    text = new char[buf.length()];
    buf.getChars(0, buf.length(), text, 0);
  }
  
  protected void compParse() throws IntercalException {
    int si, i, j, polite, total, info;
    boolean ok;
    
    Statement cmd;
    StringBuffer asm;

    ti = polite = total = 0; line = 1;
    skipSpace();
    
    while (ti < text.length) {
      cmd = new Statement();
      cmd.line = line;
      asm = cmd.asm;

      try {
        // parse an optional label
        if (text[ti] == '(') {
          cmd.label = parseLabel();
          if (labels.put(new Integer(cmd.label), cmd) != null) fail(182);
          if ((cmd.label >= SYSL_LABELMIN) && (cmd.label <= SYSL_LABELMAX)) usesyslib = SYSL_NOINCLUDE;
        }
        
        // remember where the statement starts, parse the statement identifier
        si = ti;
        ok = false;
        if (matchSkipSpace("PLEASE")) { ok = true; polite++; }
        if (matchSkipSpace("DO")) ok = true;
        if (!ok) fail(17);
        total++;

        // parse optional qualifiers
        if (matchSkipSpace("NOT") || matchSkipSpace("N'T")) info = INFO_ABSTAIN;
        else info = 0;
        
        if (text[ti] == '%') {
          int p;
          
          ti++; skipSpace();
          p = parseNumber();
          if ((p <= 0) || (p >= 100)) fail(17, cmd.line);
          
          info |= p;
        }
        else info |= 100;

        // generate statement header
        asm.append('%');
        asm.append((char) cmd.line);
        asm.append((char) info);
        asm.append((char) 0);
        asm.append((char) 0);
        
        // parse the command
        cmd.id = -1;
        
        if ((cmd.id < 0) && isVarType(text[ti])) {
          int var;
          
          cmd.id = CMD_CALCULATE;
          var = parseVarName();
          
          if ((var >> 16) >= VAR_ARRAY16) {
            if (matchSkipSpace("<-")) {
              asm.append(parseSubscripts((char) 0, true));
              asm.append(']');
              asm.append((char) (var >> 16));
              asm.append((char) (var & 0xffff));
            }
            else if (matchSkipSpace("SUB")) {
              String subscript = parseSubscripts('<', false);
              if (!matchSkipSpace("<-")) fail(17);
              asm.append(parseExpression());
              asm.append(subscript);
              asm.append((var & 0x10000) == 0 ? '-' : '=');
              asm.append((char) (var & 0xffff));
            }
            else fail(17);
          }
          else {
            if (!matchSkipSpace("<-")) fail(17);
            asm.append(parseExpression());
            asm.append((var & 0x10000) == 0 ? '-' : '=');
            asm.append((char) (var & 0xffff));
          }
        }
        
        if ((cmd.id < 0) && (text[j = ti] == '(')) {
          int label = parseLabel();
          if (matchSkipSpace("NEXT")) {
            cmd.id = CMD_NEXT;
            cmd.offset = label;
            if ((label >= SYSL_LABELMIN) && (label <= SYSL_LABELMAX) &&
                (usesyslib >= 0)) usesyslib = SYSL_INCLUDE;
            
            asm.append('n');
            asm.append((char) 0);
            asm.append((char) 0);
          }
          else ti = j; // not a NEXT, reset position
        }
        
        if ((cmd.id < 0) && matchSkipSpace("RESUME")) {
          cmd.id = CMD_RESUME;
          asm.append(parseExpression());
          asm.append('r');
        }
        
        if ((cmd.id < 0) && matchSkipSpace("FORGET")) {
          cmd.id = CMD_FORGET;
          asm.append(parseExpression());
          asm.append('f');
        }
        
        if ((cmd.id < 0) && matchSkipSpace("STASH")) {
          int[] vars;
          
          cmd.id = CMD_STASH;
          vars = parseVarList();
          
          for (j=0; j<vars.length; j++) {
            asm.append('s');
            asm.append((char) (vars[j] >> 16 | INFO_STASH));
            asm.append((char) (vars[j] & 0xffff));
          }
        }
        
        if ((cmd.id < 0) && matchSkipSpace("RETRIEVE")) {
          int[] vars;
          
          cmd.id = CMD_RETRIEVE;
          vars = parseVarList();
          
          for (j=0; j<vars.length; j++) {
            asm.append('s');
            asm.append((char) (vars[j] >> 16));
            asm.append((char) (vars[j] & 0xffff));
          }
        }
          
        if ((cmd.id < 0) && matchSkipSpace("IGNORE")) {
          int[] vars;
          
          cmd.id = CMD_IGNORE;    
          vars = parseVarList();
          
          for (j=0; j<vars.length; j++) {
            asm.append('i');
            asm.append((char) (vars[j] >> 16 | INFO_ABSTAIN));
            asm.append((char) (vars[j] & 0xffff));
          }
        }
        
        if ((cmd.id < 0) && matchSkipSpace("REMEMBER")) {
          int[] vars;
          
          cmd.id = CMD_REMEMBER;
          vars = parseVarList();
          
          for (j=0; j<vars.length; j++) {
            asm.append('i');
            asm.append(((char) (vars[j] >> 16)));
            asm.append(((char) (vars[j] & 0xffff)));
          }
        }
        
        if ((cmd.id < 0) && match2SkipSpace("ABSTAIN", "FROM")) {
          cmd.id = CMD_ABSTAIN;
          
          if (text[ti] == '(') {
            cmd.offset = parseLabel();
            asm.append('?');
            asm.append((char) INFO_ABSTAIN);
            asm.append((char) 0);
            asm.append((char) 0);
          }
          else {
            int[] ids = parseGerundList();
            for (j=0; j<ids.length; j++) {
              asm.append('!');
              asm.append((char) (INFO_ABSTAIN | ids[j]));
            }
          }
        }
        
        if ((cmd.id < 0) && matchSkipSpace("REINSTATE")) {
          cmd.id = CMD_REINSTATE;
          
          if (text[ti] == '(') {
            cmd.offset = parseLabel();
            asm.append('?');
            asm.append((char) 0);
            asm.append((char) 0);
            asm.append((char) 0);
          }
          else {
            int[] ids = parseGerundList();
            for (j=0; j<ids.length; j++) {
              asm.append('!');
              asm.append((char) ids[j]);
            }
          }
        }
        
        if ((cmd.id < 0) && match2SkipSpace("READ", "OUT")) {
          int var;
          
          cmd.id = CMD_READOUT;
          do {
            if (text[ti] == '#') {
              ti++; skipSpace();
              var = parseNumber(); if (var > 0xffff) fail(17);
              asm.append('#');
              asm.append((char) var);
              asm.append('>');
            }
            else {
              var = parseVarName();
              if ((var >> 16) >= VAR_ARRAY16) {
                if (matchSkipSpace("SUB")) {
                  asm.append(parseSubscripts('+', false));
                  asm.append((var & 0x10000) == 0 ? '.' : ':');
                  asm.append((char) (var & 0xffff));
                  asm.append('>');
                }
                else {
                  asm.append('}');
                  asm.append((char) (var >> 16));
                  asm.append((char) (var & 0xffff));
                }
              }
              else {
                asm.append((var & 0x10000) == 0 ? '.' : ':');
                asm.append((char) (var & 0xffff));
                asm.append('>');
              }
            }
          } while (matchSkipSpace("+"));
        }
        
        if ((cmd.id < 0) && match2SkipSpace("WRITE", "IN")) {
          int var;
          
          cmd.id = CMD_WRITEIN;
          do {
            var = parseVarName();
            if ((var >> 16) >= VAR_ARRAY16) {
              if (matchSkipSpace("SUB")) {
                asm.append('<');
                asm.append(parseSubscripts('+', false));
                asm.append((var & 0x10000) == 0 ? '-' : '=');
                asm.append((char) (var & 0xffff));
              }
              else {
                asm.append('{');
                asm.append((char) (var >> 16));
                asm.append((char) (var & 0xffff));
              }
            }
            else {
              asm.append('<');
              asm.append((var & 0x10000) == 0 ? '-' : '=');
              asm.append((char) (var & 0xffff));
            }
          } while (matchSkipSpace("+"));
        }
        
        if ((cmd.id < 0) && match2SkipSpace("COME", "FROM")) {
          cmd.id = CMD_COMEFROM;
          cmd.offset = parseLabel();
          asm.append('g');
          asm.append((char) 0);
          asm.append((char) 0);
        }
        
        if ((cmd.id < 0) && match2SkipSpace("GIVE", "UP")) {
          cmd.id = CMD_GIVEUP;
          
          if ((info & INFO_ABSTAIN) == 0) asm.append('E');
          else asm.setLength(0);
        }
        
        if (cmd.id < 0) {
          cmd.id = CMD_SPLAT;
          
          // find a place to restart parsing
          j = ti;
          while (ti < text.length) {
            i = ti;
            if (matchString("PLEASE")) { ti = i; break; }
            if (matchString("DO"))     { ti = i; break; }
            ti++;
          }
          if (ti < text.length) reverseSkipLabel();
          else while (isSpace(text[ti-1])) ti--;
          
          // translate '\n' into ' ' in the splat text and adjust line count
          for (i=si; i<ti; i++) {
            if (text[i] == '\n') {
              if (i >= j) line++;
              text[i] = ' ';
            }
          }
          
          // only compile splats which aren't abstained from or have a label,
          // because the others can never be reinstated
          if (((info & INFO_ABSTAIN) == 0) || (cmd.label != 0)) {
            asm.append('*');
            asm.append((char) (ti-si));
            asm.append(text, si, ti-si);
          }
          else asm.setLength(0);
          
          skipSpace();
        }
      }
      catch (IntercalException ie) { ie.setLine(cmd.line); throw ie; }
      catch (Exception e) { fail(17, cmd.line); }
      
      // update statement length, store command
      if (asm.length() > 0) {
        asm.setCharAt(INFO_CMDSIZE, (char) (asm.length() - SIZE_CMDHEADER));
        code.addElement(cmd);
      }
    }

    if (polite < total / 5) fail(79);
    if (polite > (total+2) / 3) fail(99);
    
    // clean up the globals
    text = null; ti = 0;
  }
  
  protected void compResolve() throws IntercalException {
    Statement cmd, target;
    int i;

    for (i=0; i<code.size(); i++) {
      cmd = (Statement) code.elementAt(i);
      
      if (cmd.id == CMD_COMEFROM) {
        target = (Statement) labels.get(new Integer(cmd.offset));
        if (target == null) fail(444, cmd.line);
        if (target.comefrom != null) fail(555, cmd.line);
        target.comefrom = cmd;
        target.asm.append(cmd.asm.toString());
        cmd.asm = null;
      }
      if (cmd.id == CMD_NEXT) {
        target = (Statement) labels.get(new Integer(cmd.offset));
        if (target == null) fail(129, cmd.line);
        cmd.target = target;
      }
      if (((cmd.id == CMD_ABSTAIN) || (cmd.id == CMD_REINSTATE)) && (cmd.offset != 0)) {
        target = (Statement) labels.get(new Integer(cmd.offset));
        if (target == null) fail(139, cmd.line);
        if (!(target.id <= CMD_CANABSTAIN)) fail(303, cmd.line);
        cmd.target = target;
      }
    }
  }
  
  protected void compFlatten() {
    Statement cmd;
    int i, offset = 0;
    
    for (i=0; i<code.size(); i++) {
      cmd = (Statement) code.elementAt(i);
      
      // set offset and realoffset - the offsets differ only for COME FROMs, offset points to
      // the trap location, while realoffset points to the location where the COME FROM
      // will jump to (the location of the COME FROM statement)
      if (cmd.id != CMD_COMEFROM) cmd.offset = offset;
      cmd.realoffset = offset;
      if (cmd.asm != null) offset += cmd.asm.length();
      
      // if the statement has a COME FROM attached, determine that COME FROM's offset
      if (cmd.comefrom != null) cmd.comefrom.offset = offset - (SIZE_CMDHEADER + SIZE_COMEFROM);
    }
  }

  protected void compLink() {
    Statement cmd;
    int i;
    
    for (i=0; i<code.size(); i++) {
      cmd = (Statement) code.elementAt(i);
      
      if (cmd.id == CMD_NEXT) {
        cmd.asm.setCharAt(LINK_NEXT,     (char) (cmd.target.realoffset & 0xffff));
        cmd.asm.setCharAt(LINK_NEXT + 1, (char) (cmd.target.realoffset >> 16));
      }
      if (((cmd.id == CMD_ABSTAIN) || (cmd.id == CMD_REINSTATE)) && (cmd.target != null)) {
        cmd.asm.setCharAt(LINK_ABSTAIN,     (char) (cmd.target.offset & 0xffff));
        cmd.asm.setCharAt(LINK_ABSTAIN + 1, (char) (cmd.target.offset >> 16));
      }
      if (cmd.comefrom != null) { // ...g__
        cmd.asm.setCharAt(cmd.asm.length() - 2, (char) (cmd.comefrom.realoffset & 0xffff));
        cmd.asm.setCharAt(cmd.asm.length() - 1, (char) (cmd.comefrom.realoffset >> 16));
      }
    }
  }
  
  protected void compOutput(String name, File target) throws IOException {
    StringBuffer buf = new StringBuffer(8192);
    Statement cmd;
    int[] cmdmap = new int[CMD_NUMCMDMAP];
    int i;
    
    // link statements in abstain lists
    for (i=0; i<CMD_NUMCMDMAP; i++) cmdmap[i] = -1;
    for (i=code.size()-1; i>=0; i--) {
      cmd = (Statement) code.elementAt(i);
      if (cmd.asm == null || cmd.id >= CMD_NUMCMDMAP) continue;
      
      if (cmdmap[cmd.id] >= 0)
        cmd.asm.setCharAt(INFO_DELTANEXT, (char) (cmdmap[cmd.id] - cmd.offset));
      cmdmap[cmd.id] = cmd.offset;
      
      if (cmd.comefrom != null) { // COME FROMs are special
        int offset = cmd.asm.length() - (SIZE_CMDHEADER + SIZE_COMEFROM);
        
        if (cmdmap[CMD_COMEFROM] >= 0)
          cmd.asm.setCharAt(offset + INFO_DELTANEXT,
                            (char) (cmdmap[CMD_COMEFROM] - (cmd.offset + offset)));
        cmdmap[cmd.id] = cmd.offset + offset;
      }
    }
    
    // create header
    buf.append((char) VERSION);
    for (i=0; i<4; i++) buf.append((char) maxvar[i]);
    for (i=0; i<CMD_NUMCMDMAP; i++) buf.append((char) (cmdmap[i]+1));
    
    // append code
    for (i=0; i<code.size(); i++) {
      cmd = (Statement) code.elementAt(i);
      if ((cmd != null) && (cmd.asm != null)) buf.append(cmd.asm.toString());
    }
    
    // write the file
    writeClassFile(name, target, buf.toString());
  }


  /////////////////////////////////////////////////////////////////////////////
  // utility methods - parsing
  //
  protected void skipSpace() {
    while ((ti < text.length) && isSpace(text[ti])) {
      if (text[ti] == '\n') line++;
      ti++;
    }
  }

  protected boolean matchString(String pattern) {
    if ((text.length - ti >= pattern.length()) &&
            pattern.equals(new String(text, ti, pattern.length()))) {
      ti += pattern.length();
      return true;
    }
    else return false;
  }
  
  protected boolean matchSkipSpace(String pattern) {
    if (matchString(pattern)) {
      skipSpace();
      return true;
    }
    else return false;
  }
  
  protected boolean match2SkipSpace(String a, String b) {
    int j = ti;
    if (matchSkipSpace(a) && matchSkipSpace(b)) return true;
    ti = j;
    return false;
  }
  
  protected void reverseSkipLabel() {
    int j = ti;
    
    try {
      while (isSpace(text[ti-1])) ti--; j = ti;
      if (text[ti-1] != ')') return; ti--;
      while (isSpace(text[ti-1])) ti--;
      if (!Character.isDigit(text[ti-1])) { ti = j; return; }
      while (Character.isDigit(text[ti-1])) ti--;
      while (isSpace(text[ti-1])) ti--;
      if (text[ti-1] != '(') { ti = j; return; } ti--;
      while (isSpace(text[ti-1])) ti--;
    }
    catch (ArrayIndexOutOfBoundsException e) { ti = j; }
    return;
  }
    
  protected int parseNumber() throws Exception {
    int n = 0, i = 0;
    
    while ((i++ < 5) && (ti < text.length) && Character.isDigit(text[ti])) {
      n = n*10 + Character.digit(text[ti++], 10);
      skipSpace();
    }
    if (i == 0) fail(17);
    
    return n;
  }
  
  protected int parseVarName() throws Exception {
    return parseVarName((char) 0);
  }
  
  protected int parseVarName(char type) throws Exception {
    int varname;
    
    if (type == 0) {
      type = text[ti++];
      skipSpace();
    }

    type = getVarType(type);
    varname = parseNumber();
    if ((varname == 0) || (varname > 0xffff)) fail(17);
    
    if (varname > maxvar[type]) maxvar[type] = varname;
    return (varname-1) | ((int) type) << 16;
  }
  
  protected int[] parseVarList() throws Exception {
    Vector vars;
    int[] res;
    int i;
    
    vars = new Vector(8, 8);
    while (true) {
      vars.addElement(new Integer(parseVarName()));
      if ((ti >= text.length) || (text[ti] != '+')) break;
      ti++; skipSpace();
    }
    
    res = new int[vars.size()];
    for (i=0; i<vars.size(); i++) res[i] = ((Integer) vars.elementAt(i)).intValue();
    
    return res;
  }
  
  protected int parseGerund() throws Exception {
    if (matchSkipSpace("CALCULATING"))      return CMD_CALCULATE;
    if (matchSkipSpace("NEXTING"))          return CMD_NEXT;
    if (matchSkipSpace("RESUMING"))         return CMD_RESUME;
    if (matchSkipSpace("FORGETTING"))       return CMD_FORGET;
    if (matchSkipSpace("STASHING"))         return CMD_STASH;
    if (matchSkipSpace("RETRIEVING"))       return CMD_RETRIEVE;
    if (matchSkipSpace("IGNORING"))         return CMD_IGNORE;
    if (matchSkipSpace("REMEMBERING"))      return CMD_REMEMBER;
    if (matchSkipSpace("ABSTAINING"))       return CMD_ABSTAIN;
    if (matchSkipSpace("REINSTATING"))      return CMD_REINSTATE;
    if (match2SkipSpace("READING", "OUT"))  return CMD_READOUT;
    if (match2SkipSpace("WRITING", "IN"))   return CMD_WRITEIN;
    if (match2SkipSpace("COMING", "FROM"))  return CMD_COMEFROM;
    fail(17);
    return -1;
  }

  protected int[] parseGerundList() throws Exception {
    Vector gerunds;
    int[] res;
    int i;
    
    gerunds = new Vector(8, 8);
    while (true) {
      gerunds.addElement(new Integer(parseGerund()));
      if ((ti >= text.length) || (text[ti] != '+')) break;
      ti++; skipSpace();
    }
    
    res = new int[gerunds.size()];
    for (i=0; i<gerunds.size(); i++) res[i] = ((Integer) gerunds.elementAt(i)).intValue();
    
    return res;
  }
  
  protected int parseLabel() throws Exception {
    int label;
    
    if (text[ti++] != '(') fail(17); skipSpace();
    label = parseNumber(); if ((label == 0) || (label > 0xffff)) fail(197);
    skipSpace(); if (text[ti++] != ')') fail(17);
    skipSpace();
    
    return label;
  }
  
  protected char parseUnop() {
    char unop;
    
    switch (text[ti]) {
      case '&': unop = 'a'; break;
      case 'V': unop = 'o'; break;
      case '?': unop = 'x'; break;
      default: return 0;
    }
    ti++; skipSpace();
    return unop;
  }
  
  protected String parseExpression() throws Exception {
    return parseExpression((char) 0);
  }
  
  protected String parseExpression(char delimit) throws Exception {
    StringBuffer asm = new StringBuffer(512);
    char binop;
    
    asm.append(parseUnaryExpression(delimit));
    if (ti >= text.length) return asm.toString();
    
    if ((text[ti] == '$') || (text[ti] == '~')) {
      binop = text[ti++]; skipSpace();
      asm.append(parseUnaryExpression(delimit));
      asm.append(binop);
      
      // mingle is always 32bit, select has the same size as the right operand
      if (binop == '$') expr32 = true;
    }
    
    return asm.toString();
  }
  
  protected String parseUnaryExpression(char delimit) throws Exception {
    StringBuffer asm = new StringBuffer(512);
    char type, spark, unop = 0;
    int n;
    
    switch (text[ti]) {
      case '\'':
      case '"':
        spark = text[ti++]; skipSpace();
        unop = parseUnop();
        asm.append(parseExpression(spark));
        if (text[ti++] != spark) fail(17);
        skipSpace();
        break;
        
      case '!': // wow is an abbreviation for spark-spot
        text[ti] = '.';
        asm.append(parseExpression('\''));
        if (text[ti++] != '\'') fail(17);
        skipSpace();
        break;
      
      case '#':
        ti++; skipSpace();
        unop = parseUnop();
        n = parseNumber(); if (n > 0xffff) fail(17);
        
        asm.append('#');
        asm.append(((char) n));
        expr32 = false;
        break;
        
      case '.': case ':':
      case ',': case ';':
        type = text[ti++]; skipSpace();
        unop = parseUnop();
        n = parseVarName(type);
        
        if ((n >> 16) >= VAR_ARRAY16) {
          if (!matchSkipSpace("SUB")) fail(17);
          asm.append(parseSubscripts(delimit, false));
        }
        if ((n & 0x10000) == 0) {
          asm.append('.');
          asm.append((char) (n & 0xffff));
          expr32 = false;
        }
        else {
          asm.append(':');
          asm.append((char) (n & 0xffff));
          expr32 = true;
        }
        break;
        
      default:
        fail(17);
    }
    
    if (unop != 0) {
      if (expr32) unop = Character.toUpperCase(unop);
      asm.append(unop);
    }
    
    return asm.toString();
  }
  
  protected String parseSubscripts(char delimit, boolean parseBy) throws Exception {
    StringBuffer asm = new StringBuffer(512);
    int dim = 0;
    
    do {
      asm.append(parseExpression(delimit));
      dim++;
    } while (ti < text.length &&
             !isDelimiter(delimit, text[ti]) &&
             (!parseBy || matchSkipSpace("BY")));
    
    asm.append('[');
    asm.append((char) dim);
    
    return asm.toString();
  }
  
  protected static boolean isDelimiter(char delimit, char c) {
    return (delimit != 0)
           ? c == delimit
           : c == 'D' || c == 'P' || c == '('; // DO PLEASE or (label)
  }
  
  protected static boolean isSpace(char c) {
    return (c == ' ') || (c == '\t') || (c == '\n');
  }
  
  protected static boolean isVarType(char c) {
    return (c == '.') || (c == ':') || (c == ',') || (c == ';');
  }    
  
  // hack: returns type + 1, so 0 can be used for "no match"
  protected static char getVarType(char type) throws Exception {
    switch (type) {
      case '.': return VAR_SCALAR16;
      case ':': return VAR_SCALAR32;
      case ',': return VAR_ARRAY16;
      case ';': return VAR_ARRAY32;
    }
    fail(17);
    return 0;
  }


  /////////////////////////////////////////////////////////////////////////////
  // utility methods - error handling
  //
  protected static void fail(int error, int line) throws IntercalException {
    throw new IntercalException(error, line);
  }

  protected static void fail(int error) throws IntercalException {
    throw new IntercalException(error);
  }

  protected static void fail() throws IntercalException {
    throw new IntercalException(778);
  }


  /////////////////////////////////////////////////////////////////////////////
  // utility methods - class stub handling
  //
  protected void writeClassFile(String classname, File filename, String code) throws IOException {
    DataOutputStream out = new DataOutputStream(new FileOutputStream(filename));
    if (STUB_CLASSNAME < STUB_CODE) {
      out.write(STUB, 0, STUB_CLASSNAME);
      out.writeUTF(classname);
      out.write(STUB, STUB_CLASSNAME + S_STUB_CLASSNAME.length,
                STUB_CODE - (STUB_CLASSNAME + S_STUB_CLASSNAME.length));
      out.writeUTF(code);
      out.write(STUB, STUB_CODE + S_STUB_CODE.length,
                S_STUB.length - (STUB_CODE + S_STUB_CODE.length));
    }
    else {
      out.write(STUB, 0, STUB_CODE);
      out.writeUTF(code);
      out.write(STUB, STUB_CODE + S_STUB_CODE.length,
                STUB_CLASSNAME - (STUB_CODE + S_STUB_CODE.length));
      out.writeUTF(classname);
      out.write(STUB, STUB_CLASSNAME + S_STUB_CLASSNAME.length,
                S_STUB.length - (STUB_CLASSNAME + S_STUB_CLASSNAME.length));
    }
    out.close();
  }

  protected static byte[] initConvert(short[] shorts) {
    byte[] bytes = new byte[shorts.length];
    for (int i=0; i<shorts.length; i++) bytes[i] = (byte) shorts[i];
    return bytes;
  }

  protected static int initFind(short[] data, short[] pattern) {
    int i, j;
    for (i=0; i<data.length+1-pattern.length; i++) {
      for (j=0; j<pattern.length; j++) if (data[i+j] != pattern[j]) break;
      if (j == pattern.length) return i;
    }
    return -1;
  }

  protected static short[] S_STUB = {
    0xca, 0xfe, 0xba, 0xbe, 0x00, 0x03, 0x00, 0x2d, 0x00, 0x25, 0x08, 0x00, 0x20, 0x07, 0x00, 0x1a,
    0x07, 0x00, 0x1e, 0x07, 0x00, 0x1f, 0x07, 0x00, 0x21, 0x07, 0x00, 0x22, 0x0a, 0x00, 0x06, 0x00,
    0x0b, 0x0a, 0x00, 0x04, 0x00, 0x0c, 0x0a, 0x00, 0x04, 0x00, 0x0d, 0x0a, 0x00, 0x04, 0x00, 0x0e,
    0x0c, 0x00, 0x14, 0x00, 0x0f, 0x0c, 0x00, 0x14, 0x00, 0x11, 0x0c, 0x00, 0x23, 0x00, 0x12, 0x0c,
    0x00, 0x24, 0x00, 0x0f, 0x01, 0x00, 0x03, 0x28, 0x29, 0x56, 0x01, 0x00, 0x2f, 0x28, 0x4c, 0x6a,
    0x61, 0x76, 0x61, 0x2f, 0x69, 0x6f, 0x2f, 0x49, 0x6e, 0x70, 0x75, 0x74, 0x53, 0x74, 0x72, 0x65,
    0x61, 0x6d, 0x3b, 0x4c, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x69, 0x6f, 0x2f, 0x4f, 0x75, 0x74, 0x70,
    0x75, 0x74, 0x53, 0x74, 0x72, 0x65, 0x61, 0x6d, 0x3b, 0x49, 0x29, 0x56, 0x01, 0x00, 0x41, 0x28,
    0x4c, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x6c, 0x61, 0x6e, 0x67, 0x2f, 0x53, 0x74, 0x72, 0x69, 0x6e,
    0x67, 0x3b, 0x4c, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x69, 0x6f, 0x2f, 0x49, 0x6e, 0x70, 0x75, 0x74,
    0x53, 0x74, 0x72, 0x65, 0x61, 0x6d, 0x3b, 0x4c, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x69, 0x6f, 0x2f,
    0x4f, 0x75, 0x74, 0x70, 0x75, 0x74, 0x53, 0x74, 0x72, 0x65, 0x61, 0x6d, 0x3b, 0x49, 0x29, 0x56,
    0x01, 0x00, 0x28, 0x28, 0x4c, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x6c, 0x61, 0x6e, 0x67, 0x2f, 0x53,
    0x74, 0x72, 0x69, 0x6e, 0x67, 0x3b, 0x5b, 0x4c, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x6c, 0x61, 0x6e,
    0x67, 0x2f, 0x53, 0x74, 0x72, 0x69, 0x6e, 0x67, 0x3b, 0x29, 0x56, 0x01, 0x00, 0x16, 0x28, 0x5b,
    0x4c, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x6c, 0x61, 0x6e, 0x67, 0x2f, 0x53, 0x74, 0x72, 0x69, 0x6e,
    0x67, 0x3b, 0x29, 0x56, 0x01, 0x00, 0x06, 0x3c, 0x69, 0x6e, 0x69, 0x74, 0x3e, 0x01, 0x00, 0x04,
    0x43, 0x6f, 0x64, 0x65, 0x01, 0x00, 0x0d, 0x43, 0x6f, 0x6e, 0x73, 0x74, 0x61, 0x6e, 0x74, 0x56,
    0x61, 0x6c, 0x75, 0x65, 0x01, 0x00, 0x0a, 0x45, 0x78, 0x63, 0x65, 0x70, 0x74, 0x69, 0x6f, 0x6e,
    0x73, 0x01, 0x00, 0x12, 0x4c, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x6c, 0x61, 0x6e, 0x67, 0x2f, 0x53,
    0x74, 0x72, 0x69, 0x6e, 0x67, 0x3b, 0x01, 0x00, 0x0e, 0x4c, 0x6f, 0x63, 0x61, 0x6c, 0x56, 0x61,
    0x72, 0x69, 0x61, 0x62, 0x6c, 0x65, 0x73, 0x01, 0x00, 0x0b, 0x52, 0x75, 0x6e, 0x74, 0x69, 0x6d,
    0x65, 0x53, 0x74, 0x75, 0x62, 0x01, 0x00, 0x10, 0x52, 0x75, 0x6e, 0x74, 0x69, 0x6d, 0x65, 0x53,
    0x74, 0x75, 0x62, 0x2e, 0x6a, 0x61, 0x76, 0x61, 0x01, 0x00, 0x0a, 0x53, 0x6f, 0x75, 0x72, 0x63,
    0x65, 0x46, 0x69, 0x6c, 0x65, 0x01, 0x00, 0x04, 0x63, 0x6f, 0x64, 0x65, 0x01, 0x00, 0x1a, 0x69,
    0x6e, 0x74, 0x65, 0x72, 0x63, 0x61, 0x6c, 0x2f, 0x49, 0x6e, 0x74, 0x65, 0x72, 0x63, 0x61, 0x6c,
    0x45, 0x78, 0x63, 0x65, 0x70, 0x74, 0x69, 0x6f, 0x6e, 0x01, 0x00, 0x10, 0x69, 0x6e, 0x74, 0x65,
    0x72, 0x63, 0x61, 0x6c, 0x2f, 0x52, 0x75, 0x6e, 0x74, 0x69, 0x6d, 0x65, 0x01, 0x00, 0x0c, 0x69,
    0x6e, 0x74, 0x65, 0x72, 0x63, 0x61, 0x6c, 0x63, 0x6f, 0x64, 0x65, 0x01, 0x00, 0x13, 0x6a, 0x61,
    0x76, 0x61, 0x2f, 0x69, 0x6f, 0x2f, 0x49, 0x4f, 0x45, 0x78, 0x63, 0x65, 0x70, 0x74, 0x69, 0x6f,
    0x6e, 0x01, 0x00, 0x10, 0x6a, 0x61, 0x76, 0x61, 0x2f, 0x6c, 0x61, 0x6e, 0x67, 0x2f, 0x4f, 0x62,
    0x6a, 0x65, 0x63, 0x74, 0x01, 0x00, 0x04, 0x6d, 0x61, 0x69, 0x6e, 0x01, 0x00, 0x03, 0x72, 0x75,
    0x6e, 0x00, 0x11, 0x00, 0x02, 0x00, 0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x1a, 0x00, 0x1d, 0x00,
    0x18, 0x00, 0x01, 0x00, 0x16, 0x00, 0x00, 0x00, 0x02, 0x00, 0x01, 0x00, 0x03, 0x00, 0x09, 0x00,
    0x24, 0x00, 0x10, 0x00, 0x02, 0x00, 0x15, 0x00, 0x00, 0x00, 0x1e, 0x00, 0x06, 0x00, 0x04, 0x00,
    0x00, 0x00, 0x12, 0xbb, 0x00, 0x04, 0x59, 0x12, 0x01, 0x2a, 0x2b, 0x1c, 0xb7, 0x00, 0x08, 0x4e,
    0x2d, 0xb6, 0x00, 0x0a, 0xb1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x17, 0x00, 0x00, 0x00, 0x06, 0x00,
    0x02, 0x00, 0x03, 0x00, 0x05, 0x00, 0x09, 0x00, 0x23, 0x00, 0x13, 0x00, 0x02, 0x00, 0x15, 0x00,
    0x00, 0x00, 0x13, 0x00, 0x02, 0x00, 0x01, 0x00, 0x00, 0x00, 0x07, 0x12, 0x01, 0x2a, 0xb8, 0x00,
    0x09, 0xb1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x17, 0x00, 0x00, 0x00, 0x04, 0x00, 0x01, 0x00, 0x05,
    0x00, 0x01, 0x00, 0x14, 0x00, 0x0f, 0x00, 0x01, 0x00, 0x15, 0x00, 0x00, 0x00, 0x11, 0x00, 0x01,
    0x00, 0x01, 0x00, 0x00, 0x00, 0x05, 0x2a, 0xb7, 0x00, 0x07, 0xb1, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x01, 0x00, 0x1c, 0x00, 0x00, 0x00, 0x02, 0x00, 0x1b,
  };

  protected static final short[] S_STUB_CLASSNAME = {
    0x00, 0x0b, 0x52, 0x75, 0x6e, 0x74, 0x69, 0x6d, 0x65, 0x53, 0x74, 0x75, 0x62
  };

  protected static final short[] S_STUB_CODE = {
    0x00, 0x0c, 0x69, 0x6e, 0x74, 0x65, 0x72, 0x63, 0x61, 0x6c, 0x63, 0x6f, 0x64, 0x65,
  };

  protected static final byte[] STUB        = initConvert(S_STUB);
  protected static final int STUB_CLASSNAME = initFind(S_STUB, S_STUB_CLASSNAME);
  protected static final int STUB_CODE      = initFind(S_STUB, S_STUB_CODE);
}
