// J-INTERCAL 0.12 Copyright (c) 2000-2001 Karsten Sperling, Karlsruhe, Germany.
// This software is released under the GNU LGPL, see LICENSE for details.

package intercal;

import java.io.*;
import java.util.Stack;
import java.util.Hashtable;


public final class Runtime {
  
  /////////////////////////////////////////////////////////////////////////////
  // constants
  //
  
  public final static int HINT_HELP         = 0x0001;
  public final static int HINT_WIMP         = 0x0002;
  public final static int HINT_CLOCKFACE    = 0x0004;
  public final static int HINT_DISASSEMBLE  = 0x8000;
  
  protected final static int HINT_DEFAULT   = 0;

  protected final static int VERSION        = 1;  
  protected final static int NEXTSIZE       = 79;
  protected final static int STACKSIZE      = 1024;
  
  protected final static int HDR_VERSION    = 0;
  protected final static int HDR_DATASIZE   = 1;
  protected final static int HDR_CMDLINKS   = 5;
  protected final static int HDR_HEADERSIZE = 18;
  
  protected final static int VAR_SCALAR16   = 0;
  protected final static int VAR_SCALAR32   = 1;
  protected final static int VAR_ARRAY16    = 2;
  protected final static int VAR_ARRAY32    = 3;
  
  protected final static int CMD_NUMCMDS    = 13;
  protected final static int CMD_CALCULATE  = 0;
  protected final static int CMD_NEXT       = 1;
  protected final static int CMD_RESUME     = 2;
  protected final static int CMD_FORGET     = 3;
  protected final static int CMD_STASH      = 4;
  protected final static int CMD_RETRIEVE   = 5;
  protected final static int CMD_IGNORE     = 6;
  protected final static int CMD_REMEMBER   = 7;
  protected final static int CMD_ABSTAIN    = 8;
  protected final static int CMD_REINSTATE  = 9;
  protected final static int CMD_READOUT    = 10;
  protected final static int CMD_WRITEIN    = 11;
  protected final static int CMD_COMEFROM   = 12;

  protected final static int INFO_RANDOM    = 0x00ff;
  protected final static int INFO_ABSTAIN   = 0x8000;
  protected final static int INFO_STASH     = 0x8000;
  
  protected final static int NUM_MAXDIGITS  = 10;
  protected final static int NUM_MAXROMAN   = NUM_MAXDIGITS*4;
  
  
  /////////////////////////////////////////////////////////////////////////////
  // static methods for running INTERCAL programs from the commandline
  //
  public static void main(String iclcode, String[] argv) throws IOException {
    int hints;
    Runtime rt;
    
    try {
      hints = parseArgs(argv);
      
      if ((hints & HINT_HELP) != 0) { System.err.println(
        "This is an INTERCAL program running on the JIVM step " + VERSION + ".\n\n" +
        "A set of options can be used to control the behaviour of the runtime engine.\n" +
  	    "All flags must be preceded by either + or -, which will usually mean turn\n" +
  	    "on or turn off something, but not always. For inconvenience, switches may\n" +
  	    "not be aggregated. Options currently available include:\n" +
  	    "  ?\tshow this help. Note that asking for help is\n" +
  	    "\talmost as lame as using WIMP MODE.\n" +
  	    "  c\ttoggle clockface output (e.g. use IIII instead of IV).\n" +
  	    "  w\ttoggle WIMP MODE.\n" +
  	    "  d\ttoggle output of JIVM bytecode to STDOUT.\n");
  	    return;
  	  }
      
      if ((hints & HINT_WIMP) != 0) { System.err.println(
        "How sad... you have selected to run an INTERCAL program in WIMP MODE.\n\n" +
        "This means that:\n" +
        " A) Rather than the nifty input like:\n" +
        "\tONE NINER SEVEN TWO OH SIX THREE,\n" +
        "    and even niftier output like:\n" +
        "\t______\n\tMCMLXXMMLXIII,\n" +
        "    you will have to settle for plain old number\n" +
        "    representations like 1972063; and,\n\n" +
        " B) You are a WIMP!\n");
      }
      
      rt = new Runtime(iclcode, System.in, System.out, hints);
      rt.run();
    }
    catch (IntercalException e) {
      System.err.println(e);
    }
  }
    
  protected static int parseArgs(String[] argv) throws IntercalException {
    int i, hints, option;
    char onoff;
    
    hints = HINT_DEFAULT;
    
    for (i=0; i<argv.length; i++) {
      if (argv[i].length() != 2) throw new IntercalException(601);
      
      onoff = argv[i].charAt(0);
      if ((onoff != '-') && (onoff != '+')) throw new IntercalException(601);
      
      switch (argv[i].charAt(1)) {
        case '?': option = HINT_HELP; onoff = '+'; break;
        case 'c': option = HINT_CLOCKFACE; break;
        case 'w': option = HINT_WIMP; break;
        case 'd': option = HINT_DISASSEMBLE; break;
        default: throw new IntercalException(601);
      }
      
      if (onoff == '+') hints |= option;
      else hints &= ~option;
    }
    
    return hints;
  }


  /////////////////////////////////////////////////////////////////////////////
  // intercal arrays
  //
  protected class Array16 {
    public int[] dimensions;
    public short[] data;
    
    public Array16(int[] dimensions) throws IntercalException {
      int i, size = 1;
      
      this.dimensions = new int[dimensions.length];
      for (i=0; i<dimensions.length; i++) {
        if (dimensions[i] <= 0) fail(240);
        size *= (this.dimensions[i] = dimensions[i]);
      }
      
      data = new short[size];
    }
    
    public Array16(Array16 other) {
      dimensions = new int[other.dimensions.length];
      System.arraycopy(other.dimensions, 0, dimensions, 0, dimensions.length);
      data = new short[other.data.length];
      System.arraycopy(other.data, 0, data, 0, data.length);
    }
    
    protected int getOffset(int[] subscript) throws IntercalException {
      int i, offset = 0;
      
      if (subscript.length != dimensions.length) fail(241);
      for (i=0; i<dimensions.length; i++) {
        if ((subscript[i] <= 0) || (subscript[i] > dimensions[i])) fail(241);
        if (i>0) offset *= dimensions[i-1];
        offset += subscript[i] - 1; // intercal arrays start at 1
      }
      
      return offset;
    }
    
    public short fetch(int[] subscript) throws IntercalException {
      return data[getOffset(subscript)];
    }
    
    public void store(int[] subscript, short value) throws IntercalException {
      data[getOffset(subscript)] = value;
    }
  }
    

  protected class Array32 {
    public int[] dimensions;
    public int[] data;
    
    public Array32(int[] dimensions) throws IntercalException {
      int i, size = 1;
      
      this.dimensions = new int[dimensions.length];
      for (i=0; i<dimensions.length; i++) {
        if (dimensions[i] <= 0) fail(240);
        size *= (this.dimensions[i] = dimensions[i]);
      }
      
      data = new int[size];
    }
    
    public Array32(Array32 other) {
      dimensions = new int[other.dimensions.length];
      System.arraycopy(other.dimensions, 0, dimensions, 0, dimensions.length);
      data = new int[other.data.length];
      System.arraycopy(other.data, 0, data, 0, data.length);
    }
    
    protected int getOffset(int[] subscript) throws IntercalException {
      int i, offset = 0;
      
      if (subscript.length != dimensions.length) fail(241);
      for (i=0; i<dimensions.length; i++) {
        if ((subscript[i] <= 0) || (subscript[i] > dimensions[i])) fail(241);
        if (i>0) offset *= dimensions[i-1];
        offset += subscript[i] - 1; // intercal arrays start at 1
      }
      
      return offset;
    }
    
    public int fetch(int[] subscript) throws IntercalException {
      return data[getOffset(subscript)];
    }
    
    public void store(int[] subscript, int value) throws IntercalException {
      data[getOffset(subscript)] = value;
    }
  }

  
  /////////////////////////////////////////////////////////////////////////////
  // instance methods and data, the runtime engine itself
  //
  protected InputStream in;
  protected OutputStream out;
  protected BufferedReader inw;
  protected PrintWriter outw;

  protected char[]  code;
  protected int[]   ignorance, commands;

  protected short[] var16;
  protected int[]   var32;
  protected Array16[] array16;
  protected Array32[] array32;
  
  protected int hints, line, lastin, lastout;
  

  public Runtime(String iclcode, InputStream in, OutputStream out, int hints) {
    this.in    = in;
    this.out   = out;
    this.hints = hints;
    
    inw   = null;
    outw  = null;
    code  = new char[iclcode.length()];
    iclcode.getChars(0, iclcode.length(), code, 0);

    if ((hints & HINT_DISASSEMBLE) != 0) {
      for (int i=0; i<code.length; i++)
        System.out.println("'" + code[i] + "' = " + ((int) code[i]));
    }
  }
  
  public void run() throws IntercalException, IOException {
    int[] stack, subscript, next;
    Hashtable stash;
    int ip, sp, np;
    int a, b;
    
    ///////////////////////////////////////////////////////////////////////////
    // JIVM setup
    //
    if (code[HDR_VERSION] > VERSION) fail(701);
    
    stack = new int[STACKSIZE];
    next  = new int[NEXTSIZE];
    stash = new Hashtable(128, 0.7f);
    subscript = null;
    line  = -1;
    sp = np = -1;
    ip = HDR_HEADERSIZE;
    lastin = lastout = 0;
    
    // allocate variables and ignorance map
    a = 0;
    var16   = new short  [b = code[HDR_DATASIZE + VAR_SCALAR16]]; a += b;
    var32   = new int    [b = code[HDR_DATASIZE + VAR_SCALAR32]]; a += b;
    array16 = new Array16[b = code[HDR_DATASIZE + VAR_ARRAY16]]; a += b;
    array32 = new Array32[b = code[HDR_DATASIZE + VAR_ARRAY32]]; a += b;

    ignorance = new int[(a + 31) / 32];
                         
    // command type lists
    commands = new int[CMD_NUMCMDS];
    for (b=0; b<CMD_NUMCMDS; b++) commands[b] = code[HDR_CMDLINKS + b];
    

    ///////////////////////////////////////////////////////////////////////////
    // execution
    //
    try {
      execution: while (true) {
        if (ip >= code.length) fail(633);
        
        switch (code[ip++]) {
          case 'E': // exit
            break execution;
            
          case '%': // statement [line, info, length, delta next]
            line = code[ip++];
            a    = code[ip++];
            b    = code[ip]; ip += 2;
            
            if (((a & INFO_ABSTAIN) != 0) || (((a & INFO_RANDOM) < 100) && (Math.random()*100 > (a & INFO_RANDOM))))
              ip += b;
            break;
            
          case '*': // splat [length, text...]
            a = code[ip++];
            throw new IntercalException(new String(code, ip, a), line);
            
          case 'g': // goto [target lo, target hi]
            a = ((int) code[ip]) + (((int) code[ip+1]) << 16) + HDR_HEADERSIZE; ip += 2;
            
            if (a >= code.length) fail();
            ip = a;
            break;
            
          case 'n': // next [target lo, target hi]
            a = ((int) code[ip]) + (((int) code[ip+1]) << 16) + HDR_HEADERSIZE; ip += 2;
            
            if (a >= code.length) fail();
            if (np+1 >= NEXTSIZE) fail(123);
            next[++np] = ip;
            ip = a;
            break;
            
          case 'r': // resume
            np -= stack[sp--];
            if (np < -1) fail(632);
            ip = next[np+1];
            break;
            
          case 'f': // forget
            np -= stack[sp--];
            if (np < -1) np = -1;
            break;
            
          case 's': // stash/retrieve [info, number]
            a = code[ip++];
            b = code[ip++];
            if ((a & INFO_STASH) != 0) {
              Integer key;
              Object value, previous;
              
              key = new Integer(b | ((a & 0xff) << 16));
              switch (a & 0xff) {
                case VAR_SCALAR16: value = new Integer(var16[b]); break;
                case VAR_SCALAR32: value = new Integer(var32[b]); break;
                case VAR_ARRAY16:  value = new Array16(array16[b]); break;
                case VAR_ARRAY32:  value = new Array32(array32[b]); break;
                default: value = null; fail();
              }
              
              previous = stash.get(key);
              if (previous == null) stash.put(key, value);
              else if (previous instanceof Stack) ((Stack) previous).push(value);
              else {
                Stack s = new Stack();
                s.push(previous);
                s.push(value);
                stash.put(key, s);
              }
            }
            else {
              Integer key = new Integer(b | ((a & 0xff) << 16));
              Object value = stash.get(key);
              if (value == null) fail(436);
              if (value instanceof Stack) {
                Stack s = (Stack) value;
                value = s.pop();
                if (s.empty()) stash.remove(key);
              }
              else stash.remove(key);
              
              switch (a & 0xff) {
                case VAR_SCALAR16: var16[b]   = (short) ((Integer) value).intValue(); break;
                case VAR_SCALAR32: var32[b]   = ((Integer) value).intValue(); break;
                case VAR_ARRAY16:  array16[b] = (Array16) value; break;
                case VAR_ARRAY32:  array32[b] = (Array32) value; break;
              }
            }
            break;
            
          case 'i': // ignore/remember [info, number]
            a = code[ip++];
            b = code[ip++];
            setIgnorance(a & ~INFO_ABSTAIN, b, (a & INFO_ABSTAIN) != 0);
            break;
          
          case '?': // abstain/reinstate by label [info, target lo, target hi]
            a = code[ip++];
            b = ((int) code[ip]) + (((int) code[ip+1]) << 16) + HDR_HEADERSIZE; ip += 2;

            if ((a & INFO_ABSTAIN) != 0) code[b+2] |= INFO_ABSTAIN;
            else code[b+2] &= ~INFO_ABSTAIN;
            break;

          case '!': // abstain/reinstate by gerund [info]
            a = code[ip++];
            b = a & 0xff;
            
            if ((b = commands[b]) != 0) {
              b += HDR_HEADERSIZE - 1;
              if ((a & INFO_ABSTAIN) != 0) {
                while (true) {
                  code[b+2] |= INFO_ABSTAIN;
                  if (code[b+4] == 0) break; b += code[b+4];
                }
              }
              else {
                while (true) {
                  code[b+2] &= ~INFO_ABSTAIN;
                  if (code[b+4] == 0) break; b += code[b+4];
                }
              }
            }
            break;
            
          case '#': // load constant [value]
            stack[++sp] = code[ip++];
            break;
            
          case '[': //  load subscript information [dimensions]
            a = code[ip++];
            subscript = new int[a];
            for (a--; a>=0; a--) subscript[a] = stack[sp--];
            break;
            
          case ']': // dimension array [info, number]
            a = code[ip++];
            b = code[ip++];
            if (a == VAR_ARRAY16) array16[b] = new Array16(subscript);
            else array32[b] = new Array32(subscript);
            subscript = null;
            break;
            
          case '.': // fetch (16) [number]
            a = code[ip++];
            if (subscript != null) {
              if (array16[a] == null) fail(241);
              stack[++sp] = ((int) array16[a].fetch(subscript)) & 0xffff;
            }
            else stack[++sp] = ((int) var16[a]) & 0xffff;
            subscript = null;
            break;
            
          case ':': // fetch (32) [number]
            a = code[ip++];
            if (subscript != null) {
              if (array32[a] == null) fail(241);
              stack[++sp] = array32[a].fetch(subscript);
            }
            else stack[++sp] = var32[a];
            subscript = null;
            break;
            
          case '-': // store (16) [number]
            a = code[ip++];
            b = stack[sp--]; if (b > 0xffff) fail(275);
            if (subscript != null) {
              if (array16[a] == null) fail(241);
              if (!getIgnorance(VAR_ARRAY16, a)) array16[a].store(subscript, (short) b);
            }
            else if (!getIgnorance(VAR_SCALAR16, a)) var16[a] = (short) b;
            subscript = null;
            break;
            
          case '=': // store (32) [number]
            a = code[ip++];
            b = stack[sp--];
            if (subscript != null) {
              if (array32[a] == null) fail(241);
              if (!getIgnorance(VAR_ARRAY32, a)) array32[a].store(subscript, b);
            }
            else if (!getIgnorance(VAR_SCALAR32, a)) var32[a] = b;
            subscript = null;
            break;
            
          case '$': // mingle
            b = stack[sp--];
            a = stack[sp--];
            if ((a > 0xffff) || (b > 0xffff)) fail(533);
            stack[++sp] = mingle(a, b);
            break;
            
          case '~': // select
            b = stack[sp--];
            a = stack[sp--];
            stack[++sp] = select(a, b);
            break;

          case 'a': // and (16)
            a = stack[sp--]; if (a > 0xffff) fail(275);
            stack[++sp] = rotate(a, 16) & a;
            break;
            
          case 'A': // and (32)
            a = stack[sp--];
            stack[++sp] = rotate(a, 32) & a;
            break;
            
          case 'o': // or (16)
            a = stack[sp--]; if (a > 0xffff) fail(275);
            stack[++sp] = rotate(a, 16) | a;
            break;
            
          case 'O': // or (32)
            a = stack[sp--];
            stack[++sp] = rotate(a, 32) | a;
            break;
            
          case 'x': // xor (16)
            a = stack[sp--]; if (a > 0xffff) fail(275);
            stack[++sp] = rotate(a, 16) ^ a;
            break;
            
          case 'X': // xor (32)
            a = stack[sp--];
            stack[++sp] = rotate(a, 32) ^ a;
            break;
            
          case '<': // plain scalar input
            stack[++sp] = plainInput();
            break;

          case '>': // plain scalar output
            plainOutput(stack[sp--]);
            break;
          
          case '{': // binary input [info, number]
            a = code[ip++];
            b = code[ip++];
            binaryInput((a << 16) | b);
            break;
            
          case '}': // binary output [info, number]
            a = code[ip++];
            b = code[ip++];
            binaryOutput((a << 16) | b);
            break;
            
          default:  // invalid opcode
            fail();
        }
      }
    }
    catch (NullPointerException e1)           { e1.printStackTrace(); fail(); } // DEBUG
    catch (ArrayIndexOutOfBoundsException e2) { e2.printStackTrace(); fail(); }
  }
  
  protected void fail(int error) throws IntercalException {
    throw new IntercalException(error, line);
  }
  
  protected void fail() throws IntercalException { fail(778); }
  
  protected boolean getIgnorance(int type, int id) {
    if (type > VAR_SCALAR16) id += var16.length;
    if (type > VAR_SCALAR32) id += var32.length;
    if (type > VAR_ARRAY16) id  += array16.length;
    if (type > VAR_ARRAY32) id  += array32.length;
    
    return ((ignorance[id/32] >> (id % 32)) & 1) != 0;
  }
  
  protected void setIgnorance(int type, int id, boolean onoff) {
    if (type > VAR_SCALAR16) id += var16.length;
    if (type > VAR_SCALAR32) id += var32.length;
    if (type > VAR_ARRAY16) id  += array16.length;
    if (type > VAR_ARRAY32) id  += array32.length;
    
    if (onoff) ignorance[id/32] |= 1 << (id % 32);
    else ignorance[id/32] &= ~(1 << (id % 32));
  }
    
  protected int mingle(int a, int b) {
    a = ((a & 0x0000ff00) << 8) | (a & 0x000000ff);
    a = ((a & 0x00f000f0) << 4) | (a & 0x000f000f);
    a = ((a & 0x0c0c0c0c) << 2) | (a & 0x03030303);
    a = ((a & 0x22222222) << 1) | (a & 0x11111111);
    b = ((b & 0x0000ff00) << 8) | (b & 0x000000ff);
    b = ((b & 0x00f000f0) << 4) | (b & 0x000f000f);
    b = ((b & 0x0c0c0c0c) << 2) | (b & 0x03030303);
    b = ((b & 0x22222222) << 1) | (b & 0x11111111);
    return (a << 1) | b;
  }
  
  protected int select(int a, int b) {
    int i, x;
    
    i = 1; x = 0;
    while ((a!=0) && (b!=0)) {
      if ((b & i) != 0) {
        x |= a & i;
        b ^= i;
        i <<= 1;
      }
      else {
        a >>= 1;
        b >>= 1;
      }
    }
    return x;
  }
  
  protected int rotate(int a, int bits) {
    return (a >> 1) | ((a & 1) << (bits-1));
  }
    
  protected int plainInput() throws IntercalException, IOException {
    String line;
    long l=0;
    
    if (inw == null) inw = new BufferedReader(new InputStreamReader(in), 1);
    line = inw.readLine();
    if ((line == null) || (line.length() == 0)) fail(562);
    
    if ((hints & HINT_WIMP) != 0) {
      try { l = Long.parseLong(line); }
      catch (NumberFormatException e) { fail(579); }
    }
    else {
      int i, j, cnt;
      char[] text = new char[line.length()];
      line.getChars(0, line.length(), text, 0);
      
      i = cnt = 0;
      while (true) {
        while ((i < text.length) && ((text[i] == ' ') || (text[i] == '\t'))) i++;
        if (i >= text.length) break;
        j = i; while ((j < text.length) && (text[j] != ' ') && (text[j] != '\t')) j++;
        l = l*10 + numeralToInt(line.substring(i, j));
        i = j; cnt++;
      }
      if (cnt == 0) fail(562);
    }
    
    if ((l < 0) || (l > 0xffffffffL)) fail(533);
    return (int) l;
  }
  
  protected void plainOutput(int i) throws IntercalException, IOException {
    long l;
    
    if (outw == null) outw = new PrintWriter(out);
    l = ((long) i) & 0xffffffffL;
    
    if ((hints & HINT_WIMP) != 0) outw.println(l);
    else outw.println(butcher(l, (hints & HINT_CLOCKFACE) != 0 ? 1 : 0));
    outw.flush();
  }
  
  protected void binaryInput(int var) throws IntercalException, IOException {
    if ((var >> 16) == VAR_ARRAY16) {
      Array16 target;
      byte[] buf;
      int pos, read, i;
      
      target = array16[var & 0xffff];
      if ((target == null) || (target.dimensions.length != 1)) fail(241);
      buf = new byte[target.data.length]; pos = 0;
      
      while (pos < buf.length) {
        read = in.read(buf, pos, buf.length-pos);
        if (read >= 0) pos += read;
        else break;
      }
      for (i=0; i<pos; i++) {
        target.data[i] = (short) ((buf[i] - lastin) & 0xff);
        lastin = buf[i];
      }
      while (i<buf.length) target.data[i++] = 0x100;
    }
    else {
      Array32 target;
      byte[] buf;
      int pos, read, i;
      
      target = array32[var & 0xffff];
      if ((target == null) || (target.dimensions.length != 1)) fail(241);
      buf = new byte[target.data.length]; pos = 0;
      
      while (pos < buf.length) {
        read = in.read(buf, pos, buf.length-pos);
        if (read >= 0) pos += read;
        else break;
      }
      for (i=0; i<pos; i++) {
        target.data[i] = (buf[i] - lastin) & 0xff;
        lastin = buf[i];
      }
      while (i<buf.length) target.data[i++] = 0x100;
    }
  }
  
  protected void binaryOutput(int var) throws IntercalException, IOException {
    if ((var >> 16) == VAR_ARRAY16) {
      Array16 source;
      byte[] buf;
      int i, c;
      
      source = array16[var & 0xffff];
      if ((source == null) || (source.dimensions.length != 1)) fail(241);
      buf = new byte[source.data.length];
      
      for (i=0; i<buf.length; i++) {
        lastout = c = lastout - source.data[i];
        c = (c & 0x0f) << 4 | (c & 0xf0) >> 4;
        c = (c & 0x33) << 2 | (c & 0xcc) >> 2;
        c = (c & 0x55) << 1 | (c & 0xaa) >> 1;
        buf[i] = (byte) (c & 0xff);
      }
      out.write(buf);
    }
  }
  
  protected static int numeralToInt(String numeral) throws IntercalException {
    int i;
    for (i=0; i<numWords.length; i++)
      if (numeral.equals(numWords[i])) return numNumbers[i];
    throw new IntercalException(579);
  }
  
  protected static String butcher(long value, int mode) {
    int i, j, digit, sig;
    char[] result, overbar;
    
    // zero is special
    if (value == 0) return "_\n ";
    
    result  = new char[NUM_MAXROMAN];
    overbar = new char[NUM_MAXROMAN];
    i = NUM_MAXROMAN;
    
    for (sig=0; (sig < NUM_MAXDIGITS) && (value > 0); sig++) {
      digit = (int) (value % 10); value /= 10;
      for (j=brnTrans[mode][digit][0]; j>0; j--) {
        result[--i] = brnDigit  [sig][brnTrans[mode][digit][j]];
        overbar[i]  = brnOverbar[sig][brnTrans[mode][digit][j]];
      }
    }
    
    return new String(overbar, i, NUM_MAXROMAN-i) + "\n" + new String(result, i, NUM_MAXROMAN-i);
  }
  
  protected static String[] numWords = {
    // English
    "OH", "ZERO", "ONE", "TWO", "THREE", "FOUR", "FIVE", "SIX", "SEVEN", "EIGHT", "NINE", "NINER",
    // Sanskrit
    "SUTYA", "SHUTYA", "EKA", "DVI", "TRI", "CHATUR", "PANCHAN", "SHASH", "SAPTAM", "ASHTAN", "NAVAN",
    // Basque
    "ZEROA", "BAT", "BI", "HIRO", "LAU", "BORTZ", "SEI", "ZAZPI", "ZORTZI", "BEDERATZI",
    // Tagalog
    "WALA", "ISA", "DALAWA", "TATLO", "APAT", "LIMA", "ANIM", "PITO", "WALO", "SIYAM",
    // Classical Nahuatl
    "AHTLE", "CE", "OME", "IEI", "NAUI", "NACUILI", "CHIQUACE", "CHICOME", "CHICUE", "CHICUNAUI",
    // Georgian
    "NULI", "ERTI", "ORI", "SAMI", "OTXI", "XUTI", "EKSVI", "SHVIDI", "RVA", "CXRA",
    // Kwakiutl (technically, Kwak'wala)
    "KE'YOS", "'NEM", "MAL'H", "YUDEXW", "MU", "SEK'A", "Q'ETL'A", "ETLEBU", "MALHGWENALH", "'NA'NE'MA",
    // Volap\"uk
    "NOS", "BAL", "TEL", "KIL", "FOL", "LUL", "M\\\"AL", "VEL", "J\\\"OL", "Z\\\"UL",
    // Esperanto
    "NULO", "UNU", "DU", "TRI", "KVAR", "KVIN", "SES", "SEP", "OK", "NAU", "NAUH", "NAUX", "NAU^", "NA^U",
  };
    
  protected static int[] numNumbers = {
    0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9,       // English
    0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,          // Sanskrit
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,             // Basque
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,             // Tagalog
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,             // Classical Nahuatl
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,             // Georgian
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,             // Kwakiutl (technically, Kwak'wala)
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9,             // Volap\"uk
    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 9, 9, 9, // Esperanto
  };
  
  protected static short[][][] brnTrans = {
    { // "normal" butchered mode
      { 0, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
      { 2, 0, 0, 0, 0 }, { 3, 0, 0, 0, 0 },
      { 2, 1, 2, 0, 0 }, { 1, 2, 0, 0, 0 },
      { 2, 2, 1, 0, 0 }, { 3, 2, 1, 1, 0 },
      { 4, 2, 1, 1, 1 }, { 2, 1, 3, 0, 0 },
    },
    { // clockface mode
      { 0, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
      { 2, 0, 0, 0, 0 }, { 3, 0, 0, 0, 0 },
      { 4, 0, 0, 0, 0 }, { 1, 2, 0, 0, 0 }, // <-- here's the difference
      { 2, 2, 1, 0, 0 }, { 3, 2, 1, 1, 0 },
      { 4, 2, 1, 1, 1 }, { 2, 1, 3, 0, 0 },
    },
  };
  
  protected static char[][] brnDigit = {
  	{ 'I', 'I', 'V', 'X' }, { 'X', 'X', 'L', 'C' },
  	{ 'C', 'C', 'D', 'M' }, { 'M', 'I', 'V', 'X' },
  	{ 'X', 'X', 'L', 'C' }, { 'C', 'C', 'D', 'M' },
  	{ 'M', 'i', 'v', 'x' }, { 'x', 'x', 'l', 'c' },
  	{ 'c', 'c', 'd', 'm' }, { 'm', 'i', 'v', 'x' },
  };
  
  protected static char[][] brnOverbar = {
  	{ ' ', ' ', ' ', ' ' },	{ ' ', ' ', ' ', ' ' },
  	{ ' ', ' ', ' ', ' ' },	{ ' ', '_', '_', '_' },
  	{ '_', '_', '_', '_' },	{ '_', '_', '_', '_' },
  	{ '_', ' ', ' ', ' ' },	{ ' ', ' ', ' ', ' ' },
  	{ ' ', ' ', ' ', ' ' },	{ ' ', '_', '_', '_' },
  };
}
